@@ -1,7 +1,11 @@
+from contextlib import contextmanager
+from locust.exception import (
+    RescheduleTask,
+    StopUser,
+    CatchResponseError,
+    InterruptTaskSet,
+)  # need to do this first to make sure monkey patching is done
 import gevent
-import gevent.monkey
-
-gevent.monkey.patch_all()
 import psycogreen.gevent
 import json
 
@@ -17,10 +21,10 @@ from datetime import datetime, timezone, timedelta
 
 import greenlet
 from dateutil import parser
-from locust.exception import RescheduleTask, StopUser, CatchResponseError, InterruptTaskSet
 import subprocess
 import locust.env
-from typing import List
+from typing import Callable, List
+from gevent.lock import Semaphore
 
 # pylint: disable=trailing-whitespace # pylint is confused by multiline strings used for SQL
 
@@ -39,35 +43,71 @@ class Timescale:  # pylint: disable=R0902
     See timescale_listener_ex.py for documentation
     """
 
+    dblock = Semaphore()
+    first_instance = True
+
     def __init__(
         self,
         env: locust.env.Environment,
-        testplan: str,
-        profile_name: str = "",
-        description: str = "",
+        testplan: str = None,
     ):
-        self.grafana_url = env.parsed_options.grafana_url
-        try:
-            self._conn = self._dbconn()
-        except psycopg2.OperationalError as e:
-            logging.error(e)
-            os._exit(1)
-        self._user_conn = self._dbconn()
-        self._testrun_conn = self._dbconn()
-        self._events_conn = self._dbconn()
-        assert testplan != ""
-        self._testplan = testplan
-        self._test_env = env.parsed_options.test_env
+        if not Timescale.first_instance:
+            # we should refactor this into a module as it is much more pythonic
+            raise Exception(
+                "You tried to initialize the Timescale listener twice, maybe both in your locustfile and using command line --timescale? Ignoring second initialization."
+            )
+        Timescale.first_instance = False
+        if testplan:
+            self._testplan = testplan  # legacy
+        elif env.parsed_options.override_plan_name:  # type: ignore[union-attr]
+            self._testplan = env.parsed_options.override_plan_name
+        else:
+            self._testplan = env.parsed_options.locustfile
         self.env = env
+        self._samples: List[dict] = []
+        self._background = gevent.spawn(self._run)
         self._hostname = socket.gethostname()  # pylint: disable=no-member
         self._username = os.getenv("USER", "unknown")
-        self._test_version = env.parsed_options.test_version
-        self._samples: List[dict] = []
         self._finished = False
-        self._profile_name = profile_name
-        self._rps = os.getenv("LOCUST_RPS", "0")
-        self._description = description
         self._pid = os.getpid()
+        events = self.env.events
+        events.test_start.add_listener(self.on_start)
+        events.request.add_listener(self.on_request)
+        events.cpu_warning.add_listener(self.on_cpu_warning)
+        events.quit.add_listener(self.on_quit)
+        events.spawning_complete.add_listener(self.spawning_complete)
+        atexit.register(self.log_stop_test_run)
+
+    @contextmanager
+    def dbcursor(self):
+        with self.dblock:
+            try:
+                if self.dbconn.closed:
+                    self.dbconn = self._dbconn()
+                yield self.dbconn.cursor()
+            except psycopg2.Error:
+                try:
+                    # try to recreate connection
+                    self.dbconn = self._dbconn()
+                except:
+                    pass
+                raise
+
+    def on_cpu_warning(self, environment: locust.env.Environment, cpu_usage, message=None, timestamp=None, **kwargs):
+        # passing a custom message & timestamp to the event is a haxx to allow using this event for reporting generic events
+        if not timestamp:
+            timestamp = datetime.now(timezone.utc).isoformat()
+        if not message:
+            message = f"{self._testplan} High CPU usage ({cpu_usage}%)"
+        with self.dbcursor() as cur:
+            cur.execute("INSERT INTO events (time, text) VALUES (%s, %s)", (timestamp, message))
+
+    def on_start(self, environment: locust.env.Environment):
+        try:
+            self.dbconn = self._dbconn()
+        except psycopg2.OperationalError as e:
+            logging.error(e)
+            sys.exit(1)
         try:
             self._gitrepo = subprocess.check_output(
                 "git remote show origin -n 2>/dev/null | grep h.URL | sed 's/.*://;s/.git$//' || true",
@@ -81,8 +121,8 @@ class Timescale:  # pylint: disable=R0902
 
         if is_worker() or is_master():
             # swarm generates the run id for its master and workers
-            if "LOCUST_RUN_ID" in os.environ:
-                self._run_id = parser.parse(os.environ["LOCUST_RUN_ID"])
+            if getattr(environment.parsed_options, "run_id", False):
+                self._run_id = parser.parse(environment.parsed_options.run_id)
             else:
                 logging.info(
                     "You are running distributed, but without swarm. run_id:s in Timescale will not match exactly between load gens"
@@ -93,26 +133,26 @@ class Timescale:  # pylint: disable=R0902
             self._run_id = datetime.now(timezone.utc)
         if not is_worker():
             logging.info(
-                f"Follow test run here: {self.grafana_url}&var-testplan={self._testplan}&from={int(self._run_id.timestamp()*1000)}&to=now"
+                f"Follow test run here: {self.env.parsed_options.grafana_url}&var-testplan={self._testplan}&from={int(self._run_id.timestamp()*1000)}&to=now"
             )
             self.log_start_testrun()
             self._user_count_logger = gevent.spawn(self._log_user_count)
 
-        self._background = gevent.spawn(self._run)
-        events = self.env.events
-        events.request.add_listener(self.on_request)
-        events.quitting.add_listener(self.quitting)
-        events.spawning_complete.add_listener(self.spawning_complete)
-        atexit.register(self.exit)
-
     def _dbconn(self) -> psycopg2.extensions.connection:
         try:
             conn = psycopg2.connect(
-                host=os.environ["PGHOST"], keepalives_idle=120, keepalives_interval=20, keepalives_count=6
+                host=self.env.parsed_options.pghost,
+                user=self.env.parsed_options.pguser,
+                password=self.env.parsed_options.pgpassword,
+                database=self.env.parsed_options.pgdatabase,
+                port=self.env.parsed_options.pgport,
+                keepalives_idle=120,
+                keepalives_interval=20,
+                keepalives_count=6,
             )
         except Exception:
             logging.error(
-                "Use standard postgres env vars to specify where to report locust samples (https://www.postgresql.org/docs/11/libpq-envars.html)"
+                "Could not connect to postgres. Use standard postgres env vars or --pg* command line options to specify where to report locust samples (https://www.postgresql.org/docs/13/libpq-envars.html)"
             )
             raise
         conn.autocommit = True
@@ -123,13 +163,18 @@ class Timescale:  # pylint: disable=R0902
             if self.env.runner is None:
                 return  # there is no runner, so nothing to log...
             try:
-                with self._user_conn.cursor() as cur:
+                with self.dbcursor() as cur:
                     cur.execute(
                         """INSERT INTO user_count(time, run_id, testplan, user_count) VALUES (%s, %s, %s, %s)""",
                         (datetime.now(timezone.utc), self._run_id, self._testplan, self.env.runner.user_count),
                     )
             except psycopg2.Error as error:
                 logging.error("Failed to write user count to Postgresql: " + repr(error))
+                try:
+                    # try to recreate connection
+                    self.user_conn = self._dbconn()
+                except:
+                    pass
             gevent.sleep(2.0)
 
     def _run(self):
@@ -147,26 +192,37 @@ class Timescale:  # pylint: disable=R0902
 
     def write_samples_to_db(self, samples):
         try:
-            with self._conn.cursor() as cur:
+            with self.dbcursor() as cur:
                 psycopg2.extras.execute_values(
                     cur,
-                    """INSERT INTO request(time,run_id,greenlet_id,loadgen,name,request_type,response_time,success,testplan,response_length,exception,pid,context) VALUES %s""",
+                    """INSERT INTO request(time,run_id,greenlet_id,loadgen,name,request_type,response_time,success,testplan,response_length,exception,pid,url,context) VALUES %s""",
                     samples,
-                    template="(%(time)s, %(run_id)s, %(greenlet_id)s, %(loadgen)s, %(name)s, %(request_type)s, %(response_time)s, %(success)s, %(testplan)s, %(response_length)s, %(exception)s, %(pid)s, %(context)s)",
+                    template="(%(time)s, %(run_id)s, %(greenlet_id)s, %(loadgen)s, %(name)s, %(request_type)s, %(response_time)s, %(success)s, %(testplan)s, %(response_length)s, %(exception)s, %(pid)s, %(url)s, %(context)s)",
                 )
+
         except psycopg2.Error as error:
             logging.error("Failed to write samples to Postgresql timescale database: " + repr(error))
+            sys.exit(1)
 
-    def quitting(self, **_kwargs):
+    def on_quit(self, exit_code, **kwargs):
         self._finished = True
         atexit._clear()  # make sure we dont capture additional ctrl-c:s # pylint: disable=protected-access
         self._background.join(timeout=10)
-        if not is_worker():
+        if getattr(self, "_user_count_logger", False):
             self._user_count_logger.kill()
-        self.exit()
+        self.log_stop_test_run(exit_code)
 
     def on_request(
-        self, request_type, name, response_time, response_length, exception, context, start_time=None, **_kwargs
+        self,
+        request_type,
+        name,
+        response_time,
+        response_length,
+        exception,
+        context,
+        start_time=None,
+        url=None,
+        **kwargs,
     ):
         success = 0 if exception else 1
         if start_time:
@@ -185,6 +241,7 @@ class Timescale:  # pylint: disable=R0902
             "request_type": request_type,
             "response_time": response_time,
             "success": success,
+            "url": url[0:255] if url else None,
             "testplan": self._testplan,
             "pid": self._pid,
             "context": psycopg2.extras.Json(context, safe_serialize),
@@ -209,20 +266,22 @@ class Timescale:  # pylint: disable=R0902
         self._samples.append(sample)
 
     def log_start_testrun(self):
-        with self._testrun_conn.cursor() as cur:
+        cmd = sys.argv
+        del cmd[0]
+        with self.dbcursor() as cur:
             cur.execute(
-                "INSERT INTO testrun (id, testplan, profile_name, num_clients, rps, description, env, username, gitrepo, changeset_guid) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",
+                "INSERT INTO testrun (id, testplan, num_clients, rps, description, env, username, gitrepo, changeset_guid, arguments) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",
                 (
                     self._run_id,
                     self._testplan,
-                    self._profile_name,
                     self.env.parsed_options.num_users or 1,
-                    self._rps,
-                    self._description,
-                    self._test_env,
+                    self.env.parsed_options.ips,  # this field is incorrectly called "rps" in db, it should be called something like "target_ips"
+                    self.env.parsed_options.description,
+                    self.env.parsed_options.test_env,
                     self._username,
                     self._gitrepo,
-                    self._test_version,
+                    self.env.parsed_options.test_version,
+                    " ".join(cmd),
                 ),
             )
             cur.execute(
@@ -234,21 +293,32 @@ class Timescale:  # pylint: disable=R0902
         if not is_worker():  # only log for master/standalone
             end_time = datetime.now(timezone.utc)
             try:
-                self._events_conn.cursor().execute(
-                    "INSERT INTO events (time, text) VALUES (%s, %s)",
-                    (end_time, f"{self._testplan} rampup complete, {user_count} users spawned"),
-                )
+                with self.dbcursor() as cur:
+                    cur.execute(
+                        "INSERT INTO events (time, text) VALUES (%s, %s)",
+                        (end_time, f"{self._testplan} rampup complete, {user_count} users spawned"),
+                    )
             except psycopg2.Error as error:
                 logging.error(
                     "Failed to insert rampup complete event time to Postgresql timescale database: " + repr(error)
                 )
 
-    def log_stop_test_run(self):
+    def log_stop_test_run(self, exit_code=None):
+        if is_worker():
+            return  # only run on master or standalone
+        if not self.dbconn:
+            return  # test_start never ran, so there's not much for us to do
         end_time = datetime.now(timezone.utc)
         try:
-            with self._testrun_conn.cursor() as cur:
-                cur.execute("UPDATE testrun SET end_time = %s where id = %s", (end_time, self._run_id))
-                cur.execute("INSERT INTO events (time, text) VALUES (%s, %s)", (end_time, self._testplan + " finished"))
+            with self.dbcursor() as cur:
+                cur.execute(
+                    "UPDATE testrun SET end_time = %s, exit_code = %s where id = %s",
+                    (end_time, exit_code, self._run_id),
+                )
+                cur.execute(
+                    "INSERT INTO events (time, text) VALUES (%s, %s)",
+                    (end_time, self._testplan + f" finished with exit code: {exit_code}"),
+                )
                 # The AND time > run_id clause in the following statements are there to help Timescale performance
                 # We dont use start_time / end_time to calculate RPS, instead we use the time between the actual first and last request
                 # (as this is a more accurate measurement of the actual test)
@@ -280,13 +350,9 @@ WHERE id = %s""",
                 + repr(error)
             )
         logging.info(
-            f"Report: {self.grafana_url}&var-testplan={self._testplan}&from={int(self._run_id.timestamp()*1000)}&to={int((end_time.timestamp()+1)*1000)}\n"
+            f"Report: {self.env.parsed_options.grafana_url}&var-testplan={self._testplan}&from={int(self._run_id.timestamp()*1000)}&to={int((end_time.timestamp()+1)*1000)}\n"
         )
 
-    def exit(self):
-        if not is_worker():  # on master or standalone locust run
-            self.log_stop_test_run()
-
 
 class Print:
     """
@@ -304,7 +370,7 @@ class Print:
         )
 
     def on_request(
-        self, request_type, name, response_time, response_length, exception, context: dict, start_time=None, **_kwargs
+        self, request_type, name, response_time, response_length, exception, context: dict, start_time=None, **kwargs
     ):
         if exception:
             if isinstance(exception, CatchResponseError):
@@ -351,9 +417,9 @@ class RescheduleTaskOnFail:
         # causing other listeners to be skipped
         env.events.request.add_listener(self.request)
 
-    def request(self, exception, **_kwargs):
+    def request(self, exception, **kwargs):
         if exception:
-            raise RescheduleTask()
+            raise RescheduleTask(exception)
 
 
 class InterruptTaskOnFail:
@@ -362,7 +428,7 @@ class InterruptTaskOnFail:
         # causing other listeners to be skipped
         env.events.request.add_listener(self.request)
 
-    def request(self, exception, **_kwargs):
+    def request(self, exception, **kwargs):
         if exception:
             raise InterruptTaskSet()
 
@@ -373,7 +439,7 @@ class StopUserOnFail:
         # causing other listeners to be skipped
         env.events.request.add_listener(self.request)
 
-    def request(self, exception, **_kwargs):
+    def request(self, exception, **kwargs):
         if exception:
             raise StopUser()
 
@@ -384,10 +450,10 @@ class ExitOnFail:
         # causing other listeners to be skipped
         env.events.request.add_listener(self.request)
 
-    def request(self, exception, **_kwargs):
+    def request(self, exception, **kwargs):
         if exception:
             gevent.sleep(0.2)  # wait for other listeners output to flush / write to db
-            os._exit(1)
+            sys.exit(1)
 
 
 class QuitOnFail:
@@ -398,25 +464,37 @@ class QuitOnFail:
         self.env = env
         env.events.request.add_listener(self.request)
 
-    def request(self, exception, name, **_kwargs):
+    def request(self, exception, name, **kwargs):
         if exception and (name == self.name or not self.name):
             gevent.sleep(0.2)  # wait for other listeners output to flush / write to db
             self.env.runner.quit()
 
 
+class RunOnFail:
+    def __init__(self, env: locust.env.Environment, function: Callable):
+        # execute the provided function on failure
+        self.function = function
+        env.events.request.add_listener(self.request)
+
+    def request(self, exception, **kwargs):
+        if exception:
+            self.function(exception, **kwargs)
+
+
+class RunOnUserError:
+    def __init__(self, env: locust.env.Environment, function: Callable):
+        # execute the provided function on unhandled exception in a task
+        self.function = function
+        env.events.user_error.add_listener(self.user_error)
+
+    def user_error(self, user_instance, exception, tb, **kwargs):
+        if exception:
+            self.function(user_instance, exception, tb, **kwargs)
+
+
 def is_worker():
     return "--worker" in sys.argv
 
 
 def is_master():
     return "--master" in sys.argv
-
-
-class TimescaleListener:
-    def __init__(self, *args, **kwargs):
-        raise Exception("All listeners have had their -Listener suffix removed, please update your code.")
-
-
-class RescheduleTaskOnFailListener:
-    def __init__(self, *args, **kwargs):
-        raise Exception("All listeners have had their -Listener suffix removed, please update your code.")
