@@ -6,6 +6,9 @@ import time
 from contextlib import contextmanager
 import os
 from abc import ABC, abstractmethod
+from gevent.lock import Semaphore
+
+dblock = Semaphore()
 
 
 class NoUserException(Exception):
@@ -15,9 +18,10 @@ class NoUserException(Exception):
 class User(dict):
     def __init__(self, coll: pymongo.collection.Collection, query: dict):
         self.coll = coll
-        data = self.coll.find_one_and_update(
-            query, {"$set": {"last_login": datetime.utcnow(), "logged_in": True}}, sort=[("last_login", 1)]
-        )
+        with dblock:
+            data = self.coll.find_one_and_update(
+                query, {"$set": {"last_login": datetime.utcnow(), "logged_in": True}}, sort=[("last_login", 1)]
+            )
         if not data:
             raise NoUserException(f"Didnt get any user from db ({self.coll}) using query {query}")
         super().__init__(data)
@@ -26,11 +30,15 @@ class User(dict):
         super().__setitem__(key, value)
         self.coll.find_one_and_update({"_id": self["_id"]}, {"$set": {key: value}})
 
+    def __delitem__(self, key):
+        super().__delitem__(key)
+        self.coll.find_one_and_update({"_id": self["_id"]}, {"$unset": {key: 1}})
+
 
 class Reader(ABC):
     @abstractmethod
     @contextmanager
-    def user(self):
+    def user(self, query: dict = None):
         pass
 
 
@@ -46,9 +54,15 @@ class MongoReader(Reader):
         self.query = {"$and": filters + [{"logged_in": False}]}
 
     @contextmanager
-    def user(self):
+    def user(self, query: dict = None):
         start_at = time.monotonic()
-        user = User(self.coll, self.query)
+
+        if not query:
+            if not self.query:
+                raise Exception("no query specified on object or as parameter :(")
+            query = self.query
+
+        user = User(self.coll, query)
 
         if start_at + self.delay_warning < time.monotonic():
             if not self.delay_warning:
@@ -56,7 +70,7 @@ class MongoReader(Reader):
                 self.delay_warning = 1
             else:
                 logging.warning(
-                    f"Getting a user with filter more than {self.delay_warning} seconds (doubling warning threshold for next time, filter used was {self.filters})"
+                    f"Getting a user took more than {self.delay_warning} seconds (doubling warning threshold for next time, query used was {query})"
                 )
                 self.delay_warning *= 2
         try:
